from contextlib import contextmanager
import csv
from math import floor, isnan
from statistics import mean
from typing import Any, Dict, Optional, Tuple

import cv2
import numpy as np

__all__ = [
    'extract_samples', 'locate_aruco_markers', 'read_jazirrad_file'
]


@contextmanager
def _coerce_to_2dim(x: Dict):
    '''
    This provides a context inside of which the image in a data dictionary has
    its spatial and (if present) temporal dimensions flattened into a single
    dimension:

    .. code-block:: python

        x = {'image': np.arange(96).reshape(4, 4, 2, 3)}

        with _coerce_to_2dim(x):
            print(x.shape)  # returns (32, 3)

        print(x.shape)  # returns (4, 4, 2, 3)

    This is typically used for math operations, such as linear algebra
    operations, that expect 1- or 2-dimensional inputs.
    '''
    *head, c = x['image'].shape
    x['image'] = x['image'].reshape(-1, c)
    if 'mask' in x:
        x['mask'] = x['mask'].flatten()
    yield
    x['image'] = x['image'].reshape(*head, -1)
    if 'mask' in x:
        x['mask'] = x['mask'].reshape(*head[:2])


@contextmanager
def _coerce_to_3dim(x: Dict):
    '''
    This provides a context inside of which the image in a data dictionary has
    its channel and (if present) temporal dimensions flattened into a single
    dimension:

    .. code-block:: python

        x = {'image': np.arange(96).reshape(4, 4, 2, 3)

        with _coerce_to_3dim(x):
            print(x.shape)  # returns (4, 4, 6)

        print(x.shape)  # returns (4, 4, 2, 3)

    This is typically used for vision operations that expect 3-dimensional
    inputs.
    '''
    tail = x['image'].shape[2:]
    x['image'] = x['image'].reshape(*x['image'].shape[:2], -1)
    yield
    x['image'] = x['image'].reshape(*x['image'].shape[:2], *tail)


@contextmanager
def _coerce_to_4dim(x: Dict):
    '''
    This provides a context inside of which the image in a data dictionary is
    coerced to (H, W, T, C) arrangement, adding a temporal dimension if
    necessary.

    .. code-block:: python

        x = {'image': np.arange(48).reshape(4, 4, 3)}

        with _coerce_to_2dim(x):
            print(x.shape)  # returns (4, 4, 1, 3)

        print(x.shape)  # returns (4, 4, 3)

    This is typically used for operations where we want to be able to index
    over time without worrying about whether time exists to be indexed over.
    '''
    has_t, shape = (x['image'].ndim == 4), x['image'].shape
    if not has_t:
        x['image'] = x['image'].reshape(*shape[:2], 1, shape[2])
    yield
    if not has_t:
        x['image'] = x['image'].squeeze(2)


def _coerce_to_dict(x: Any) -> Dict:
    '''
    This coerces an arbitrary object into a dictionary suitable for use in an
    :class:`Operator`.
    '''
    if not isinstance(x, dict):
        return {'image': _coerce_to_image(x)}
    x['image'] = _coerce_to_image(x['image'])
    if 'mask' in x:
        x['mask'] = _coerce_to_mask(x['mask'])
    return x


def _coerce_to_image(image: Any) -> np.ndarray:
    '''
    Convenience function to coerce an arbitrary objects to a
    :class:`numpy.ndarray` that is either 3- or 4-dimensional.
    '''
    image = np.array(image)
    # Assume a 2-dimensional image is an image with a single band.
    if image.ndim == 2:
        image = image.reshape(*image.shape, 1)
    elif image.ndim not in {3, 4}:
        raise ValueError(
            f'Image must be 3- or 4-dimensional, not {image.shape}'
        )
    if (image.dtype != np.float32):
        image = image.astype(np.float32)
    return image


def _coerce_to_mask(mask: Any) -> np.ndarray:
    '''
    Convenience function to coerce an arbitrary objects to a mask.
    '''
    mask = np.array(mask)
    # Assume a 2-dimensional image is an image with a single band.
    if mask.ndim != 2:
        raise ValueError(
            f'Mask must be 2-dimensional, not {mask.shape}'
        )
    if mask.dtype == np.bool_:
        mask = mask.astype(np.uint8)
    elif mask.dtype != np.uint8:
        raise ValueError(f'mask dtype {mask.dtype} not supported')
    return mask


def extract_samples(image: np.ndarray, points: np.ndarray, width: int = 10) \
        -> np.ndarray:
    '''
    Extracts samples surrounding sample points from an image.

    Args:
        image (:class:`numpy.ndarray`): Image to extract samples from.
        points (:class:`numpy.ndarray`): Points to take samples from. These
        should be either (x, y) points if the image is an image, or (x, y, t)
        if it is a video.
        width (int): The width of the sample regions.
    '''
    image = _coerce_to_image(image)
    points = np.array(points).astype(np.int64)
    samples = np.empty((points.shape[0], image.shape[-1]), dtype=np.float32)

    min_pts, max_pts = points.min(0), points.max(0)
    (h, w, *t) = image.shape[:-1]
    if (min_pts < 0).any() or (max_pts >= (w, h, *t)).any():
        raise ValueError(
            f'Sample point outside of bounds: {points} vs {image.shape}'
        )
    if (image.ndim == 4) != (points.shape[1] == 3):
        raise ValueError(
            f'Mismatch between image shape and number of point coordinates: '
            f'{image.shape} vs {points}'
        )

    # Convert from center_x, center_y, (t) to l_x, u_y, r_x, d_y, (t).
    r = floor(width / 2)
    points = np.concatenate((
        np.maximum(points[:, 0:2] - r, 0),
        points[:, 0:2] + (width - r),
        points[:, 2:]
    ), axis=1)

    if image.ndim == 3:
        for i, (x_1, y_1, x_2, y_2) in enumerate(points):
            samples[i, :] = image[y_1:y_2, x_1:x_2, :].mean((0, 1))
    elif image.ndim == 4:
        for i, (x_1, y_1, x_2, y_2, t) in enumerate(points):
            samples[i, :] = image[y_1:y_2, x_1:x_2, t, :].mean((0, 1))

    return samples


def _evaluate_ecc_for_warp(template: Dict, image: Dict) -> float:
    '''
    Evaluates a warp by applying it to a stack of images then calculating the
    mean ECC of the warped images against a set of templates. This is a
    convenience wrapper around the function :func:`cv2.computeECC`, which only
    takes one grayscale image at a time.

    Args:
        template (dict): Template images.
        image (dict): Images to match against template images.
    '''
    with _coerce_to_4dim(template), _coerce_to_4dim(image):
        # TODO: Check template for mask also?
        mask = image.get('mask')
        image = image['image'].mean(-1, dtype=np.float32)
        template = template['image'].mean(-1, dtype=np.float32)

        eccs = []
        for t in range(template.shape[2]):
            ecc = cv2.computeECC(
                template[:, :, t],
                image[:, :, t],
                mask,
            )
            # This computation can return NaN in rare cases, e.g. if the image
            # is entirely one value.
            if not isnan(ecc):
                eccs.append(ecc)

    return mean(eccs)


def locate_aruco_markers(x: Dict, marker_ids: Optional[np.ndarray] = None):
    '''
    Convenience function for locating ARUCO markers in a batch of images.

    Something to be aware of: the ARUCO library is not able to detect flipped
    markers.

    Args:
        x (dict): The data dictionary to locate marker IDs in.
        marker_ids (optional, :class:`numpy.ndarray`): If provided, look for
        these specific marker IDs.

    Returns:
        ts (:class:`numpy.ndarray`): The time index for the frames in which the
        markers were found.
        corners: The point locations of the corners. If marker_ids is provided,
        this is returned in shape (time, marker_id, corner_id, x or y). If it
        is not, this is a list of :class:`numpy.ndarray`, each entry in shape
        (marker_id, corner_id, x or y).
        marker_ids (list of :class:`numpy.ndarray`): This is only returned if
        marker_ids is not provided. It contains the IDs of the markers.
    '''
    x = _coerce_to_dict(x)
    ts, corners, found_ids = [], [], []

    detector_params = cv2.aruco.DetectorParameters_create()
    detector_params.adaptiveThreshWinSizeMax = 256

    with _coerce_to_4dim(x):
        w = x['image'].shape[1]
        # ARUCO detector requires uint8, but we now work in float32.
        image = np.clip(x['image'] * 256., 0, 255).astype(np.uint8)

        for t in range(x['image'].shape[2]):
            pts, ids, _ = cv2.aruco.detectMarkers(
                image[:, :, t, :],
                _MARKER_DICTIONARY,
                parameters=detector_params,
            )
            # The ARUCO detector checks for rotated markers, but not flipped.
            # So we do it manually.
            if ids is None:
                pts, ids, _ = cv2.aruco.detectMarkers(
                    image[:, ::-1, t, :],
                    _MARKER_DICTIONARY,
                    parameters=detector_params,
                )
                for pt in pts:
                    pt[:, :, 0] = w - pt[:, :, 0]
            if ids is not None:
                ts.append(t)
                pts = np.stack([box.reshape(4, 2) for box in pts], axis=0)
                corners.append(pts)
                found_ids.append(list(ids.flatten()))

    if not ts:
        raise RuntimeError('Failed to locate markers')

    if marker_ids is not None:
        try:
            ts, corners, found_ids = zip(*(
                (t, b, ids) for (t, b, ids) in zip(ts, corners, found_ids)
                if set(ids) == set(marker_ids)
            ))
        except ValueError:
            raise RuntimeError('Failed to locate markers')

        ts = np.array(ts)
        corners = [
            np.array([c[ids.index(i)] for i in marker_ids])
            for c, ids in zip(corners, found_ids)
        ]
        corners = np.stack(corners, axis=0)
        return ts, corners

    else:
        found_ids = [np.array(ids) for ids in found_ids]
        return np.array(ts), corners, found_ids


def read_jazirrad_file(path: str) -> Tuple[np.ndarray, np.ndarray]:
    '''
    Reads a jazirrad file containing irradiance data, and returns a pair of
    :class:`numpy.ndarray`, one containing the wavelengths and the other
    containing the irradiance.

    Args:
        path (str): Path to the file.
    '''
    with open(path, 'r') as irrad_file:
        contents = irrad_file.readlines()

    if contents[0] != 'Jaz Absolute Irradiance File\n':
        raise ValueError('File does not match Jaz Irradiance format')

    # These lines delineate the tab-separated values data containing the
    # irradiances. There's also a header we skip.
    start_idx = contents.index('>>>>>Begin Processed Spectral Data<<<<<\n') + 2
    end_idx = contents.index('>>>>>End Processed Spectral Data<<<<<\n')
    contents = contents[start_idx:end_idx]
    # contents is now a tab-separated with four columns. The 0th is wavelength
    # and the second is the spectral response in uW / cm^2 nm.
    contents = list(csv.reader(contents, dialect='excel-tab'))
    wavelengths, _, response, _ = zip(*contents)
    wavelengths = np.array([float(w) for w in wavelengths], dtype=np.float32)
    response = np.array([float(r) for r in response], dtype=np.float32)

    return wavelengths, response


class Registry:
    '''
    This provides a convenient way to create global dictionaries that map names
    to object classes. It's inspired by:

        https://github.com/facebookresearch/fvcore/blob/main/fvcore/common/registry.py
    '''
    def __init__(self):
        self._obj_name: Dict[str, Any] = {}

    def __contains__(self, name: str) -> bool:
        return name in self._obj_name

    def _do_register(self, name: str, obj: Any):
        if name in self._obj_name:
            raise ValueError(f'{name} already registered')
        self._obj_name[name] = obj

    def get(self, name: str) -> Any:
        if name not in self._obj_name:
            raise KeyError(f'{name} not found in registry')
        return self._obj_name[name]

    def register(self, obj: Any = None):
        # In this case, it's used as a decorator.
        if obj is None:
            def deco(cls: Any):
                self._do_register(cls.__name__, cls)
                return cls
            return deco

        else:
            self._do_register(obj.__name__, obj)
            return obj


_MARKER_DICTIONARY = cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_6X6_50)



from functools import partial
from math import log
from typing import Callable, Dict, List, Optional, Tuple, Union
import warnings

import numpy as np
from scipy.optimize import curve_fit, OptimizeWarning

from .operators import Operator, OPERATOR_REGISTRY
from .utils import _coerce_to_2dim

__all__ = ['build_linearizer', 'ElementwiseOperator', 'Polynomial', 'PowerLaw']


class ElementwiseOperator(Operator):
    '''
    This is a superclass for an :class:`Operator` that applies an elementwise
    operation.
    '''
    def __init__(self, funcs: List[Callable]):
        '''
        Args:
            funcs (list of callables): The functions to apply, listed per band.
        '''
        self.funcs = funcs

    def apply(self, x: Dict) -> Dict:
        if len(self.funcs) != x['image'].shape[-1]:
            raise ValueError(
                f"Number of functions ({len(self.funcs)}) does not match shape"
                f" of image ({x['image'].shape})"
            )

        with _coerce_to_2dim(x):
            out = [
                func(x['image'][..., band].flatten()).astype(np.float32)
                for band, func in enumerate(self.funcs)
            ]
            x['image'] = np.stack(out, axis=1)

        return x

    def apply_points(self, pts: np.ndarray) -> np.ndarray:
        return pts

    def apply_values(self, values: np.ndarray) -> np.ndarray:
        '''
        This is a convenience function that applies the functions to a set of
        values of shape (N, B), where N indexs the value and B indexs the band.
        '''
        return np.stack([f(s) for f, s in zip(self.funcs, values.T)], axis=1)


@OPERATOR_REGISTRY.register
class Polynomial(ElementwiseOperator):
    '''
    This applies a polynomial to each band, e.g. for linearization.
    '''
    def __init__(self,
                 polys: List[Union[np.polynomial.Polynomial, List[float]]]):
        '''
        Args:
            polys (list of :class:`numpy.polynomial.Polynomial`): The
            polynomials to apply.
        '''
        # Need to be able to pass the polynomials as coefficients to support
        # deserialization from JSON.
        if isinstance(polys[0], (np.ndarray, list, tuple)):
            polys = [np.polynomial.Polynomial(coef) for coef in polys]
        # Coerce to np.float32
        for poly in polys:
            poly.coef = poly.coef.astype(np.float32)

        super().__init__(funcs=polys)

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'polys': [p.coef.tolist() for p in self.funcs],
        }

    @classmethod
    def fit(cls, input_samples: np.ndarray, target_samples: np.ndarray,
            order: int = 3) -> np.polynomial.Polynomial:
        '''
        Fits a polynomial to a given set of inputs and desired outputs.
        '''
        # NOTE: Domain and window need to be specified explicitly or else the
        # polynomial does this internal scaling thing that makes it hard to
        # interpret the coefficients...
        return np.polynomial.Polynomial.fit(
            input_samples,
            target_samples,
            order,
            domain=[0, 255],
            window=[0, 255],
        )


class PowerLawFormula:
    '''
    This encapsulates a power law:

    .. math::

        f(x) = cb^x + d

    There is an optional shoulder component, that replaces part of the low end
    with a linear fit:

    .. math:

        f(x) = \\left{\\{\\begin{array}{ll}
            cb^x + d & x > x_0 \\\\
            c(\\lob b)b^{x_0}(x - x_0) + cb^{x_0} & x \\leq x_0
        \\end{array}\\right}
    '''
    def __init__(self, scale: float, base: float, shift: float,
                 shoulder: Optional[float] = None):
        '''
        Args:
            scale (float): Multiplier of the exponent.

            base (float): Base of the exponent.

            shift (float): Added to the result.

            shoulder (optional, float): Location of the shoulder.
        '''
        self.scale, self.base, self.shift = scale, base, shift
        self.shoulder = shoulder

    def __call__(self, x: np.ndarray) -> np.ndarray:
        # Calculate power law component. We do these operations in place
        # where possible to minimize memory use.
        pow_out = self.base**x
        pow_out *= self.scale
        pow_out += self.shift
        # Calculate linear component
        if self.shoulder is not None:
            cbx = self.scale * self.base**self.shoulder
            slope = cbx * log(self.base)
            intercept = cbx + self.shift - (self.shoulder * slope)
            linear_out = x * slope
            linear_out += intercept
            return np.where(x > self.shoulder, pow_out, linear_out)
        else:
            return pow_out

    def __repr__(self) -> str:
        return f'PowerLawFormula(scale={self.scale}, base={self.base}, ' \
               f'shift={self.shift}, shoulder={self.shoulder})'

    @classmethod
    def fit(cls, input_samples: np.ndarray, target_samples: np.ndarray,
            init_coeffs: Optional[List[float]] = None):
        '''
        Fits a power law to a given set of inputs and desired outputs.
        '''
        use_shoulder = (init_coeffs is not None) and (len(init_coeffs) == 4)

        # f needs to have explicit coefficients, no *coeffs, because the
        # curve_fit function uses signature inspection to figure out how many
        # coefficients there are.
        if use_shoulder:
            def f(x, c_0, c_1, c_2, c_3):
                return cls(c_0, c_1, c_2, c_3)(x).astype(np.float64)
            bounds = ((-np.inf, 0, -np.inf, -np.inf),
                      (np.inf, np.inf, np.inf, np.inf))
        else:
            def f(x, c_0, c_1, c_2):
                return cls(c_0, c_1, c_2, None)(x).astype(np.float64)
            bounds = ((-np.inf, 0, -np.inf,), (np.inf, np.inf, np.inf))

        # Inputs to and outputs of f need to be float64, or else optimization
        # may return incorrect results; see:
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html

        # Also, this will raise an OptimizeWarning, because the covariance
        # cannot be estimated. We catch and suppress that.

        with warnings.catch_warnings():
            warnings.simplefilter('ignore', OptimizeWarning)
            # We add explicit constraints to the coefficient values to keep the
            # curve_fit algorithm from jumping into negatives.
            coeffs, _ = curve_fit(
                f,
                input_samples.T.astype(np.float64),
                target_samples.T.astype(np.float64),
                init_coeffs,
                bounds=bounds,
            )

        return cls(*coeffs)


@OPERATOR_REGISTRY.register
class PowerLaw(ElementwiseOperator):
    '''
    This applies a power law to each band, e.g. for linearization:

    .. math::
        c \\cdot b^x + d
    '''
    def __init__(self, coefs: List[Union[PowerLawFormula, List[float]]]):
        '''
        Args:
            funcs (list of list of floats): The coefficients to apply. Each
            entry is expected to be a trple of scale, base, shift.
        '''
        if isinstance(coefs[0], (np.ndarray, list, tuple)):
            coefs = [PowerLawFormula(*coef) for coef in coefs]

        super().__init__(funcs=coefs)

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'coefs': [(f.scale, f.base, f.shift) for f in self.funcs],
        }

    @classmethod
    def fit(cls, input_samples: np.ndarray, target_samples: np.ndarray,
            init_coeffs: Optional[Tuple[float, float, float]] = None) \
            -> PowerLawFormula:
        '''
        Fits a power law to a given set of inputs and desired outputs.
        '''
        return PowerLawFormula.fit(input_samples, target_samples, init_coeffs)


def build_linearizer(input_samples: np.ndarray, target_samples: np.ndarray,
                     method: str = 'poly', order: int = 3,
                     init_coeffs: Optional[Tuple[float, float, float]] = None)\
        -> ElementwiseOperator:
    '''
    Builds an :class:`ElementwiseOperator` that acts as a linearizer. That is,
    given a set of samples received by the sensors, and a set of known target
    values, this fits a function to each band that attempts to map the
    input samples to the target samples, with the aim of undoing whatever
    post-processing occurred in the camera.

    Args:
        input_samples (:class:`numpy.ndarray`): The samples from the
        sensor. This should be of shape (num_samples, num_bands).

        target_samples (:class:`numpy.ndarray`): The expected values for
        the samples. This should be of shape (num_samples, num_bands).

        method (str): The method to use. Choices: 'poly', 'power'.

        order (int): The order of the polynomial, if that method is being used.

        init_coeffs (optional, 4-tuple of float): The initial estimates for the
        coefficients used in power law fitting.
    '''
    input_samples = np.array(input_samples)
    target_samples = np.array(target_samples)
    if input_samples.shape != target_samples.shape:
        raise ValueError(
            f'Mismatch in sample shapes: {input_samples.shape} vs '
            f'{target_samples.shape}'
        )
    if input_samples.ndim != 2:
        raise ValueError(
            f'Samples must be 2-dimensional, not {input_samples.shape}'
        )

    n_bands = input_samples.shape[1]

    if method == 'poly':
        fit, cls = partial(Polynomial.fit, order=order), Polynomial
    elif method == 'power':
        fit, cls = partial(PowerLaw.fit, init_coeffs=init_coeffs), PowerLaw
    else:
        raise ValueError(f'Did not recognize method {method}')

    funcs = [
        fit(input_samples[:, band], target_samples[:, band])
        for band in range(n_bands)
    ]

    return cls(funcs)



'''
This provides various :class:`Operator` s, which represent operations performed
on an image. We build them as distinct classes so that they can be saved to
keep track of what operations were performed to produce an image, and reused on
other images.
'''
from copy import copy
import json
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np

from .utils import _coerce_to_3dim, _coerce_to_dict, Registry

__all__ = [
    'Operator', 'ConcatenateOnBands', 'HorizontalFlip', 'LinearMap',
    'load_operator', 'OPERATOR_REGISTRY', 'Pad', 'Resize', 'Rotate',
    'TemporalShift', 'Translation', 'VerticalFlip', 'Warp'
]

# This provides a registry of operators. This is used when saving and restoring
# Operators, since it allows us to map from operator name to the operator
# class.
OPERATOR_REGISTRY = Registry()


class Operator:
    '''
    This is the base class for an operator. To add a new operator, subclass
    this class, follow the API described below, and register the operator with:

    .. code-block:: python

        import video2vision as v2v

        v2v.OPERATOR_REGISTRY.register
        class NewOperator(Operator):
            ...

    All subclasses of :class:`Operator` must provide the following methods or
    attributes:

        * __init__: All arguments must be able to be taken in a form that is
          JSON-serializable. For example, if one of the arguments is expected
          to be a :class:`numpy.ndarray`, the method must also except a list of
          floats, and coerce that list.
        * apply: This should take in one or more images or videos without any
          other arguments, apply the operator, then return the output. The
          inputs are provided as dictionaries with the following key-value
          pairs:

             - 'image': Contains a :class:`numpy.ndarray` of dtype float32 that
               is either 3-dimensional (in H, W, C arrangement) or
               4-dimensional (in H, W, T, C arrangement). This is a mandatory
               key. The image must be scaled to the range [0, 1].
             - 'mask': Contains a :class:`numpy-ndarray` of dtype uint8 that is
               2-dimensional (in H, W arrangement) that provides a boolean mask
               of which pixels are valid. This is an optional key; if not
               present, all pixels are considered valid. Can also be None, to
               denote all pixels valid.

          There should be a single output, consisting of a dictionary with the
          same structure. None of the input :class:`numpy.ndarray` s should be
          modified in-place, since they may be being used as inputs to multiple
          :class:`Operator` s. However, the dictionary may be modified in-
          place.
        * apply_points: This should map a 2-dimensional :class:`numpy.ndarray`
          of points in an input image to points in an output image. If this
          cannot be consistently done, it should raise a RuntimeError.
        * _to_json: This should serialize the operator to JSON format. The
          output should be a dictionary with the key 'class', whose value is
          the name of the operator class as a string. All other key-value pairs
          should be the keyword arguments passed to the operator when it is
          instantiated, coerced to forms that can be JSON-serialized. For
          example, a :class:`numpy.ndarray` must have the :meth:`tolist` method
          called.
        * num_inputs: This should be an attribute specifying the number of
          inputs that are expected by the apply method. If the value is -1,
          then the :class:`Operator` takes a variable number of inputs.

    All operators can be applied to either an image or a video. Images are
    expected to be a :class:`numpy.ndarray` in (H, W, C) arrangement, where H
    indexs height, W indexs width, and C indexs channel. Videos are expected to
    be a :class:`numpy.ndarray` in (H, W, T, C) arrangement, where T indexs
    time. Note that one consequence of this is that we need to be careful in
    our indexing.
    '''
    num_inputs = 1

    def __call__(self, *xs) -> Dict:
        # copy performs a shallow copy. That is, if x is a dictionary, it
        # creates a new dictionary, but the values are all the same as the
        # original values. This avoids copying images - which would be wasteful
        # of both memory and compute - but allows the apply method to modify
        # the dictionary in-place.
        xs = tuple(copy(_coerce_to_dict(x)) for x in xs)
        return self.apply(*xs)

    def __repr__(self) -> str:
        desc = self._to_json()
        name = desc.pop('class')
        return f"{name}({', '.join(f'{k}={v}' for k, v in desc.items())})"

    def _to_json(self) -> Dict:
        return {'class': self.__class__.__name__}

    def apply(self, *xs) -> Dict:
        raise NotImplementedError()

    def apply_points(self, pts: np.ndarray) -> np.ndarray:
        raise RuntimeError()


@OPERATOR_REGISTRY.register
class ConcatenateOnBands(Operator):
    '''
    This concatenates multiple images together, subselecting the bands.
    '''
    def __init__(self, bands: List[List[int]]):
        '''
        Args:
            bands (list of list of int): The bands to use.
        '''
        self.bands = bands

    def apply(self, *xs) -> Dict:
        if self.bands is None:
            self.bands = [list(range(x['image'].shape[-1])) for x in xs]

        if len(xs) != len(self.bands):
            raise ValueError(
                f'Mismatch between number of input images and merge: '
                f'{len(xs)} vs {len(self.bands)}'
            )

        # Construct holder for return value. rtn['image'] will initially be a
        # List[np.ndarray], which we will concatenate at the end.
        rtn = {'image': [], 'names': []}

        for x, bands_from_image in zip(xs, self.bands):
            image = x.pop('image')
            for band in bands_from_image:
                if band >= image.shape[-1]:
                    raise ValueError(
                        f'Band {band} not in image of shape {image.shape}'
                    )
                rtn['image'].append(image[..., band])
            # Combine masks through intersection
            if 'mask' in x:
                if 'mask' in rtn:
                    rtn['mask'] &= x.pop('mask')
                else:
                    rtn['mask'] = x.pop('mask')
            # Combine names
            if 'names' in x:
                rtn['names'].append(x.pop('names'))
            # Append any other keys
            for k, v in x.items():
                if k not in rtn:
                    rtn[k] = v
                elif rtn[k] != v:
                    raise ValueError(
                        f'Conflicting values for key {k} in concatenated '
                        f'images: {v} vs.{rtn[k]}'
                    )

        rtn['image'] = np.stack(rtn['image'], axis=-1)

        if rtn['names']:
            rtn['names'] = ['-'.join(x) for x in zip(*rtn['names'])]
        else:
            rtn.pop('names')

        return rtn

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'bands': self.bands,
        }

    def apply_points(self, pts: np.ndarray) -> np.ndarray:
        return pts

    @property
    def num_inputs(self) -> int:
        return len(self.bands)

    @property
    def num_out_bands(self) -> int:
        '''
        Number of output bands.
        '''
        return sum(len(b) for b in self.bands)


@OPERATOR_REGISTRY.register
class HorizontalFlip(Operator):
    '''
    This flips the image horizontally.
    '''
    def apply(self, x: Dict) -> Dict:
        x['image'] = x['image'][:, ::-1]
        if x.get('mask') is not None:
            x['mask'] = x['mask'][:, ::-1]
        return x


@OPERATOR_REGISTRY.register
class LinearMap(Operator):
    '''
    Applies a linear map to an image:

    .. code-block:: python

        out[y, x, band] = sum(image[y, x, :] * mat[:, band])
    '''
    def __init__(self, mat: np.ndarray):
        '''
        Args:
            mat (:class:`numpy.ndarray`): Matrix defining the linear map.
        '''
        self.mat = np.array(mat).astype(np.float32)

    def apply(self, x: Dict) -> Dict:
        if x['image'].shape[-1] != self.mat.shape[0]:
            raise ValueError(
                f"Mismatch between number of image bands and map: "
                f"{x['image'].shape[0]} vs. {self.mat.shape[0]}"
            )

        x['image'] = np.dot(x['image'], self.mat)

        return x

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'mat': self.mat.tolist(),
        }

    def apply_points(self, pts: np.ndarray) -> np.ndarray:
        return pts

    def apply_values(self, values: np.ndarray) -> np.ndarray:
        '''
        This is a convenience function that applies the functions to a set of
        values of shape (N, B), where N indexs the value and B indexs the band.
        '''
        return np.dot(values, self.mat)

    @classmethod
    def build_sensor_convertor(cls, reflectances: np.ndarray,
                               source_sensitivities: np.ndarray,
                               control_sensitivities: np.ndarray,
                               source_illum: Optional[np.ndarray] = None,
                               control_illum: Optional[np.ndarray] = None):
        '''
        Builds a :class:`LinearMap` for converting from one sensor to another
        sensor.

        Args:
            reflectances (:class:`np.ndarray`): Expected to be a 2-dimensional
            array, with the 0th dimension indexing the wavelength of the
            incident radiation and the 1st dimension indexing the material.
            source_sensitivities (:class:`np.ndarray`): Expected to be a 2-
            dimensional array, with the 0th dimension indexing the wavelength
            and the 1st dimension indexing the band.
            control_sensitivities (:class:`np.ndarray`): Expected to be a 2-
            dimensional array, with the 0th dimension indexing the wavelength
            and the 1st dimension indexing the band.
            source_illum (optional, :class:`np.ndarray`): Expected to be a 1-
            dimensional array, indexed by wavelength.
            control_illum (optional, :class:`np.ndarray`): Expected to be a 1-
            dimensional array, indexed by wavelength.
        '''
        source_illum = 1 if (source_illum is None) else source_illum
        control_illum = 1 if (control_illum is None) else control_illum

        # Predicted received QC for source and control sensor from reflectances
        source_qc = reflectances.T.dot(source_illum * source_sensitivities)
        control_qc = reflectances.T.dot(control_illum * control_sensitivities)
        # rcond must be specified to suppress a warning
        mat, *_ = np.linalg.lstsq(source_qc, control_qc, rcond=None)

        return cls(mat)


@OPERATOR_REGISTRY.register
class Pad(Operator):
    '''
    This pads an image.
    '''
    def __init__(self, pad: Tuple[int, int, int, int],
                 value: Union[int, float]):
        '''
        Args:
            pad (tuple of int): Values to pad on top, bottom, left, and right.

            value (int or float): Value to fill the padded region with.
        '''
        if isinstance(pad, int):
            pad = (pad, pad, pad, pad)
        self.pad = pad
        self.value = value

    def apply(self, x: Dict) -> Dict:
        with _coerce_to_3dim(x):
            if x.get('mask') is None:
                x['mask'] = np.ones(x['image'].shape[:2], dtype=np.uint8)
            x['image'] = cv2.copyMakeBorder(
                x['image'], *self.pad, cv2.BORDER_CONSTANT, value=self.value
            )
            x['mask'] = cv2.copyMakeBorder(
                x['mask'], *self.pad, cv2.BORDER_CONSTANT, value=0
            )

        return x

    def apply_points(self, pts: np.ndarray) -> np.ndarray:
        shift = np.array([[self.pad[2], self.pad[0]]])
        return pts + shift

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'pad': self.pad,
            'value': self.value,
        }


@OPERATOR_REGISTRY.register
class Resize(Operator):
    '''
    This class wraps the :func:`cv2.resize` function.
    '''
    def __init__(self, scale: Union[float, Tuple[int, int]],
                 sampling_mode: int = cv2.INTER_LINEAR):
        '''
        Args:
            scale (float or tuple of int): This is either a float, giving the
            proportional scale, or a pair of ints, giving the desired output
            size in pixels.

            sampling_mode (int): Method of interpolation.
        '''
        self.scale = scale
        self.sampling_mode = sampling_mode

    def apply(self, x: Dict) -> Dict:
        if isinstance(self.scale, float):
            out_h = int(self.scale * x['image'].shape[0])
            out_w = int(self.scale * x['image'].shape[1])
        else:
            out_w, out_h = self.scale

        # If this is a video, we convert it to (H, W, C * T) arrangement. cv2
        # will treat this as an image with an unusually large number of
        # channels and be perfectly fine with it.
        with _coerce_to_3dim(x):
            x['image'] = cv2.resize(
                x['image'], (out_w, out_h), interpolation=self.sampling_mode
            )
            if x.get('mask') is not None:
                x['mask'] = cv2.resize(
                    x['mask'], (out_w, out_h), interpolation=cv2.INTER_NEAREST
                )

        return x

    def apply_points(self, pts: np.ndarray) -> np.ndarray:
        if isinstance(self.scale, float):
            return self.scale * pts
        else:
            raise RuntimeError()

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'scale': self.scale,
            'sampling_mode': self.sampling_mode,
        }


@OPERATOR_REGISTRY.register
class TemporalShift(Operator):
    '''
    Shifts a video forward in time.
    '''
    def __init__(self, shift: int):
        '''
        Args:
            shift (int): The number of frames to shift. If positive, shifts the
            video forward; if negative, shifts it backward.
        '''
        if shift < 0:
            raise ValueError('Shift must be positive')
        self.shift = shift
        self.buff = None

    def apply(self, x: Dict) -> Dict:
        if self.shift == 0:
            return x

        if x['image'].ndim != 4:
            raise ValueError(
                'TemporalShift cannot be applied to single images'
            )

        w, h, t, c = x['image'].shape

        if t <= self.shift:
            raise ValueError(
                'Temporal shift requires shift to be smaller than batch size'
            )

        # We do a copy because the slice returns a view, keeping a hidden
        # reference to the original image array which will prevent it from
        # being garbage-collected.
        buff = self.buff
        self.buff = x['image'][:, :, -self.shift:, :].copy()
        if buff is not None:
            x['image'] = np.concatenate(
                (buff, x['image'][:, :, :-self.shift, :]), axis=2
            )
        else:
            pad = ((0, 0), (0, 0), (self.shift, 0), (0, 0))
            x['image'] = np.pad(x['image'][:, :, :-self.shift], pad)

        return x

    def apply_points(self, pts: np.ndarray) -> np.ndarray:
        return pts

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'shift': self.shift,
        }


@OPERATOR_REGISTRY.register
class VerticalFlip(Operator):
    '''
    This flips the image vertically.
    '''
    def apply(self, x: Dict) -> Dict:
        x['image'] = x['image'][::-1]
        if x.get('mask') is not None:
            x['mask'] = x['mask'][::-1]
        return x


@OPERATOR_REGISTRY.register
class Warp(Operator):
    '''
    This class wraps the :func:`cv2.warpAffine` and :func:`cv2.warpPerspective`
    functions.
    '''
    def __init__(self, coe: np.ndarray, output_size: Tuple[int, int],
                 sampling_mode: int = cv2.INTER_LINEAR):
        '''
        Args:
            coe (:class:`np.ndarray`): Transform matrix.

            output_size (tuple of int): The height and width of the output
            image.

            sampling_mode (int): Interpolation mode.
        '''
        self.coe = np.array(coe).astype(np.float32)
        if self.coe.shape not in {(2, 3), (3, 3)}:
            raise ValueError(
                f'Transform matrix shape not valid: {self.coe.shape}'
            )
        self.output_size = output_size
        self.sampling_mode = sampling_mode

    def apply(self, x: Dict) -> Dict:
        warp = (cv2.warpAffine if self.is_affine else cv2.warpPerspective)

        # If this is a video, we convert it to (H, W, C * T) arrangement. cv2
        # will treat this as an image with an unusually large number of
        # channels and be perfectly fine with it.
        with _coerce_to_3dim(x):
            if x.get('mask') is None:
                x['mask'] = np.ones(x['image'].shape[:2], dtype=np.uint8)
            # Note: output_size needs to be a tuple; some versions of OpenCV
            # will raise a SystemError if it's not.
            x['image'] = warp(
                x['image'], self.coe, tuple(self.output_size),
                flags=self.sampling_mode,
            )
            x['mask'] = warp(
                x['mask'], self.coe, tuple(self.output_size),
                flags=cv2.INTER_NEAREST,
            )

        return x

    def apply_points(self, pts: np.ndarray) -> np.ndarray:
        # For unknown reasons, cv2.perspectiveTransform requires the points to
        # be in (H, W, X/Y) arrangement. They also require float32 or 64.
        pts = pts.astype(np.float32).reshape(-1, 1, 2)
        if self.coe.shape[0] == 2:
            pts = cv2.transform(pts, self.coe)
        else:
            pts = cv2.perspectiveTransform(pts, self.coe)
        return pts.reshape(-1, 2)

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'coe': self.coe.tolist(),
            'output_size': self.output_size,
            'sampling_mode': self.sampling_mode,
        }

    @classmethod
    def build_from_tiepoints(cls, source_points: np.ndarray,
                             control_points: np.ndarray,
                             control_size: Tuple[int, int],
                             method: int = cv2.RANSAC) -> Operator:
        '''
        Builds a :class:`Warp` operator from a set of tiepoints between the
        image to be warped and a target image.

        Args:
            source_points (:class:`np.ndarray`): Tie points in the image to be
            warped.
            control_points (:class:`np.ndarray`): Points to warp the tie points
            to.
            control_size (tuple of int): Size of the output image.
            method (int): Method to use in warping.
        '''
        source_points = np.array(source_points)
        control_points = np.array(control_points)
        coe, _ = cv2.findHomography(source_points, control_points, method)
        return Warp(coe, control_size)

    @property
    def is_affine(self) -> bool:
        return self.coe.shape == (2, 3)


@OPERATOR_REGISTRY.register
class Rotate(Warp):
    '''
    This class performs a rotation.
    '''
    def __init__(self, angle: float, output_size: Tuple[int, int],
                 center: Optional[Tuple[int, int]] = None):
        '''
        Args:
            angle (float): Angle of the rotation.
            output_size (tuple of int): The height and width of the output
            image.
            center (optional, tuple of int): The center point in the input
            image to rotate around.
        '''
        # Note: copied from imutils to ensure consistency with older versions.
        if center is None:
            center = (output_size[1] // 2, output_size[0] // 2)
        self.angle, self.center = angle, center
        # Note: center needs to be a tuple; some versions of OpenCV will raise
        # a SystemError if it's not.
        coe = cv2.getRotationMatrix2D(tuple(center), angle, scale=1)
        super().__init__(coe=coe, output_size=output_size)

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'angle': self.angle,
            'output_size': self.output_size,
            'center': self.center,
        }


@OPERATOR_REGISTRY.register
class Translation(Warp):
    '''
    This class performs a translation.
    '''
    def __init__(self, shift_x: float, shift_y: float,
                 output_size: Tuple[int, int]):
        '''
        Args:
            shift_x (float): Shift in x coordinate.
            shift_y (float): Shift in y coordinate.
            output_size (tuple of int): The height and width of the output
            image.
        '''
        self.shift_x, self.shift_y = shift_x, shift_y
        coe = np.array([[1., 0., shift_x], [0., 1., shift_y]])
        super().__init__(coe=coe, output_size=output_size)

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'shift_x': self.shift_x,
            'shift_y': self.shift_y,
            'output_size': self.output_size,
        }


def load_operator(x: Union[str, Dict]) -> Operator:
    '''
    Loads an operator from a saved JSON file.
    '''
    if isinstance(x, str):
        with open(x, 'r') as op_file:
            x = json.load(op_file)

    op_cls = OPERATOR_REGISTRY.get(x.pop('class'))
    return op_cls(**x)



from contextlib import contextmanager
from glob import glob
import os
from typing import Dict, Iterator, List, Optional, Union

import cv2
import numpy as np

try:
    import tifffile
    has_tiff = True
except ImportError:
    has_tiff = False

try:
    import rawpy
    has_rawpy = True
except ImportError:
    has_rawpy = False

from .operators import Operator, OPERATOR_REGISTRY
from .utils import _coerce_to_image

__all__ = ['load', 'Loader', 'OutOfInputs', 'save', 'Writer']

# Default video codec
_VIDEO_CODEC = cv2.VideoWriter_fourcc(*'mp4v')

_VIDEO_FPS = 23.976023976023978

# This is a global variable used to track whether the Loader/Writer should
# read/write from disk or to an internal buffer. This internal buffer is used
# to apply a Pipeline to images in memory instead of running the normal way.
_READ_WRITE_FROM_TO_BUFFER = False

_IMAGE_EXTENSIONS = ['arw', 'jpeg', 'jpg', 'mp4', 'png', 'raw', 'tif', 'tiff']


def load(path: str) -> np.ndarray:
    '''
    Convenience function for loading images from disk. Dispatches to
    appropriate backend.

    Args:
        path (str): Path to image to load.
    '''
    # Some of the function calls below don't raise an error if the file doesn't
    # exist, they just fail silently. So let's check explicitly.
    if not os.path.exists(path):
        raise FileNotFoundError(path)

    if path.lower().endswith(('.tif', '.tiff')):
        if not has_tiff:
            raise ImportError('tifffile is needed to read tif files')
        image = tifffile.imread(path)

    elif path.lower().endswith('.mp4'):
        reader = cv2.VideoCapture(path)
        frames, ret = [], True
        while ret and reader.isOpened():
            ret, frame = reader.read()
            if ret:
                frames.append(frame)
        reader.release()
        image = np.stack(frames, axis=2)

    elif path.lower().endswith('.arw'):
        if not has_rawpy:
            raise ImportError('rawpy is needed to read arw files')
        with rawpy.imread(path) as raw_file:
            image = raw_file.postprocess()

    else:
        image = cv2.imread(path, cv2.IMREAD_UNCHANGED)

    # Rescale to [0, 1] and float32
    image = image.astype(np.float32) / 256.

    return _coerce_to_image(image)


def save(image: np.ndarray, path: str):
    '''
    Convenience function for saving images to disk. Dispatches to appropriate
    backend.

    Args:
        image (:class:`np.ndarray`): Image to save.

        path (str): Path to save image to.
    '''
    # Rescale from [0, 1] -> [0, 256]
    image = 256 * image

    if path.lower().endswith('.mp4') and image.ndim == 4:
        # MP4 only likes uint8
        image = np.clip(image, 0, 255).astype(np.uint8)
        # Arguments are path, codec, frame rate, size, is_color
        writer = cv2.VideoWriter(
            path,
            _VIDEO_CODEC,
            _VIDEO_FPS,
            image.shape[:2][::-1],
            (image.shape[-1] > 1),
        )
        for t in range(image.shape[2]):
            writer.write(image[:, :, t, :])
        writer.release()

    elif image.ndim == 4:
        raise NotImplementedError('Video is supported only for mp4')

    elif path.lower().endswith(('.tif', '.tiff')):
        if not has_tiff:
            raise ImportError('tifffile is needed to write tif files')
        # photometric='minisblack' suppresses a DeprecationWarning.
        tifffile.imwrite(path, image, photometric='minisblack')

    else:
        return cv2.imwrite(path, image)


@OPERATOR_REGISTRY.register
class Loader(Operator):
    '''
    This is an operator that loads images from disk, acting as a root of the
    :class:`Pipeline` DAG. It can be used by:

    .. code-block:: python

        loader = Loader(path)
        image = loader()

    It will raise a :class:`OutOfInputs` exception when it runs out of images
    to load. It returns images in (H, W, C) arrangement if the batch size is 1,
    and returns a stack of images in (H, W, T, C) arrangement if the batch size
    is larger than 1.
    '''
    num_inputs = 0

    def __init__(self, path: Optional[Union[str, Iterator[str]]] = None,
                 batch_size: int = 1):
        '''
        Args:
            path (str or iterator of str): Path(s) to the images to load. This
            can be a directory, in which case everything in the directory will
            be loaded, or it can be a path to a single image, or it can be a
            string with wildcards.

            batch_size (int): Number of images to return at a time.
        '''
        self.set_path(path)
        self.batch_size = batch_size
        # This is used to provide external inputs from memory to a pipeline,
        # instead of loading from disk. It is only consulted if the global
        # variable _READ_WRITE_FROM_BUFFER is true.
        self.buff: List[np.ndarray] = []

    def __iter__(self) -> Iterator[np.ndarray]:
        # If _READ_WRITE_FROM_BUFFER, we are providing inputs from memory, not
        # loading from disk. This is used to apply Pipelines to images in
        # memory.
        if _READ_WRITE_FROM_TO_BUFFER:
            while self.buff:
                yield self.buff.pop(0), None

        else:
            for path, reader in zip(self.paths, self._readers):
                name, _ = os.path.splitext(os.path.basename(path))
                if reader is not None:
                    ret, frame = reader.read()
                    while ret and reader.isOpened():
                        # Rescale to [0, 1] before returning
                        yield frame.astype(np.float32) / 256., name
                        ret, frame = reader.read()
                else:
                    # load handles rescaling for us
                    yield load(path), name

    def __len__(self) -> int:
        if _READ_WRITE_FROM_TO_BUFFER:
            return len(self.buff)
        else:
            return sum(int(_get_num_frames(r)) for r in self._readers)

    def apply(self) -> Dict:
        # If _READ_WRITE_FROM_BUFFER, we are providing inputs from memory, not
        # loading from disk. This is used to apply Pipelines to images in
        # memory.
        if _READ_WRITE_FROM_TO_BUFFER:
            if self.buff:
                out, self.buff = np.stack(self.buff, axis=2), []
                return {'image': out}
            else:
                raise OutOfInputs('Buffer is empty')

        try:
            frame, name = next(self._data_iter)
        except StopIteration:
            raise OutOfInputs()
        names = [name]

        if self.batch_size == 1:
            return {'image': frame, 'names': names}

        else:
            h, w, c = frame.shape
            rtn = np.empty((h, w, self.batch_size, c), dtype=frame.dtype)
            rtn[:, :, 0, :] = frame

            try:
                for t in range(1, self.batch_size):
                    rtn[:, :, t, :], name = next(self._data_iter)
                    names.append(name)
            except StopIteration:
                rtn = rtn[:, :, :t, :]

            return {'image': rtn, 'names': names}

    def get_frame(self, t: int) -> np.ndarray:
        '''
        Retrieves a single frame and returns it.
        '''
        if _READ_WRITE_FROM_TO_BUFFER:
            return self.buff[t]

        n_frames = 0
        for path, reader in zip(self.paths, self._readers):
            n_frames += _get_num_frames(reader)
            if n_frames > t:
                break
        else:
            raise ValueError(f'{t} out of range: {len(self)}')

        if reader is None:
            # load handles rescaling for us
            image = load(path)
        else:
            reader.set(cv2.CAP_PROP_POS_FRAMES, t)
            _, image = reader.read()
            image = image.astype(np.float32) / 256.

        return image

    def set_path(self, paths: Optional[Union[str, Iterator[str]]]):
        '''
        Sets the input path.
        '''
        if paths is None:
            self.paths, self._readers = None, []
        else:
            self.paths, self._readers = [], []
            if isinstance(paths, str):
                paths = [paths]

            for path in paths:
                if os.path.isdir(path):
                    path = os.path.join(path, '*')
                self.paths += sorted(glob(path))

            for path in self.paths:
                if path.lower().endswith('.mp4'):
                    self._readers.append(cv2.VideoCapture(path))
                else:
                    self._readers.append(None)

        self._data_iter = iter(self)

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'batch_size': self.batch_size,
        }


@OPERATOR_REGISTRY.register
class Writer(Operator):
    '''
    This is an :class:`Operator` that writes finished images to disk. It acts
    as a leaf of the :class:`Pipeline` DAG.
    '''

    def __init__(self, path: Optional[str] = None,
                 extension: Optional[str] = None,
                 separate_bands: bool = False, suffix: str = ''):
        '''
        Args:
            path (str): Path to the directory to write images to.
            extension (optional, str): Extension to use in writing images. If
            not set, attempts to infer from path; if unsuccessful, assumes
            'tif'.

            separate_bands (bool): Whether to save each band as a separate
            output file, instead of a single output file.

            suffix (str): Append this to the end of output file names. This is
            only used if the output path is a directory, not an MP4.
        '''
        if extension is None:
            # Check for extension in path
            _, extension = os.path.splitext(path or '')
            extension = extension[1:].lower()
            if extension not in _IMAGE_EXTENSIONS:
                extension = 'tif'
        if extension.lower() in {'arw', 'raw'}:
            raise NotImplementedError('Writer does not support RAW format')
        self.extension = extension
        # This will be used to hold cv2.VideoWriter if present
        self._writer = None
        self.separate_bands = separate_bands
        self.set_path(path)
        self.suffix = suffix
        # This is used to return inputs to memory from a pipeline, instead of
        # writing to disk. It is only used if the global variable
        # _READ_WRITE_FROM_BUFFER is true.
        self.buff = []

    def apply(self, x: Dict):
        image = x['image']
        num_frames = 1 if (image.ndim == 3) else image.shape[2]
        names = x.get('names', [None] * num_frames)

        if num_frames != len(names):
            raise RuntimeError(
                f'Mismatch between number of frames {num_frames} and number of'
                f' names {names}'
            )

        if image.ndim == 3:
            self._write(image, names[0])
        else:
            for t in range(image.shape[2]):
                self._write(image[:, :, t, :], names[t])

    def __del__(self):
        self.release()

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'extension': self.extension,
            'separate_bands': self.separate_bands,
        }

    def _write(self, image: np.ndarray, name: str):
        '''
        This is a convenience wrapper for writing images to disk.
        '''
        if _READ_WRITE_FROM_TO_BUFFER:
            self.buff.append(image)
            return

        if self.path.lower().endswith(self.extension.lower()):
            path = self.path
        else:
            name = f'{name}{self.suffix}.{self.extension}'
            path = os.path.join(self.path, name)

        # TODO: This should be handled better.
        if self.extension.lower() == 'mp4':
            # Rescale to [0, 256], and MP4 only likes uint8
            image = np.clip(256 * image, 0, 255).astype(np.uint8)

            if self.separate_bands:
                if self._writer is None:
                    paths = _get_band_paths(path, image.shape[2])
                    if any(os.path.exists(p) for p in paths):
                        raise FileExistsError(paths)
                    self._writer = [
                        _get_writer(p, image[..., 0:1]) for p in paths
                    ]
                if len(self._writer) != image.shape[2]:
                    raise RuntimeError(
                        f'Mismatch in number of bands: {len(self._writer)} '
                        f'vs {image.shape[2]}'
                    )
                for b, writer in enumerate(self._writer):
                    writer.write(image[..., b])
            else:
                if self._writer is None:
                    self._writer = [_get_writer(path, image)]
                self._writer[0].write(image)

        else:
            # save handles rescaling for us
            if self.separate_bands:
                paths = _get_band_paths(path, image.shape[2])
                for b, path in enumerate(paths):
                    if os.path.exists(path):
                        raise FileExistsError(path)
                    save(image[..., b], path)
            else:
                if os.path.exists(path):
                    raise FileExistsError(path)
                save(image, path)

    def release(self):
        '''
        Releases the writer after you're finished with it. This has no effect
        if you're writing individual frames at a time, but is necessary when
        writing video.
        '''
        if self._writer is not None:
            for writer in self._writer:
                writer.release()

    def set_path(self, path: Optional[str]):
        '''
        Sets the output path.
        '''
        if path is not None:
            if path.lower().endswith(self.extension.lower()):
                root = os.path.dirname(path)
            else:
                root = path
            os.makedirs(root, exist_ok=True)

        self.path = path


class OutOfInputs(Exception):
    '''
    This exception is raised by a :class:`Loader` when it exhausts its
    available inputs.
    '''


def _get_band_paths(path: str, num_bands: int) -> List[str]:
    '''
    Convenience function for creating paths for separate bands from a single
    path.
    '''
    base_path, ext = os.path.splitext(path)
    return [f'{base_path}_{b}{ext}' for b in range(num_bands)]


def _get_num_frames(r: Optional[cv2.VideoCapture]) -> int:
    '''
    Convenience function for calculating number of frames in an MP4.
    '''
    if r is None:
        return 1
    else:
        return r.get(cv2.CAP_PROP_FRAME_COUNT)


def _get_writer(path: str, image: np.ndarray) -> cv2.VideoWriter:
    '''
    Convenience function for building an MP4 writer.
    '''
    return cv2.VideoWriter(
        path,
        _VIDEO_CODEC,
        _VIDEO_FPS,
        image.shape[:2][::-1],
        (image.shape[-1] > 1),
    )


@contextmanager
def _read_write_from_to_buffer():
    '''
    Context manager inside of which :class:`video2vision.Loader` and
    :class:`video2vision.Writer` will read from and write to internal buffers instead of disk.
    '''
    global _READ_WRITE_FROM_TO_BUFFER
    _READ_WRITE_FROM_TO_BUFFER, was = True, _READ_WRITE_FROM_TO_BUFFER
    try:
        yield
    finally:
        _READ_WRITE_FROM_TO_BUFFER = was



from collections import Counter
from contextlib import suppress
import json
import time
from typing import Dict, List, Optional, Union

import networkx as nx
import numpy as np

from .io import Loader, OutOfInputs, _read_write_from_to_buffer, Writer
from .operators import Operator, OPERATOR_REGISTRY

__all__ = ['load_pipeline', 'Pipeline']


class Pipeline(nx.DiGraph):
    '''
    :class:`Pipeline` instantiates a full image processing pipeline. It
    subclasses :class:`networkx.DiGraph`, which represents a directed graph.
    Each node in the graph represents an image processing :class:`Operator`.
    Each directed edge in the graph represents the output of one operator being
    passed to another operator as input.

    Since networkx may not be familiar to everyone, we provide detailed notes
    on the parts of :class:`networkx.DiGraph` we are using. First: each node in
    the graph must be represented by a hashable object. We use integer indices.
    The nodes property returns an iterator over all of the nodes. In our case,
    that means it iterates over the integer indices, not the underlying
    :class:`Operator` s. In addition, a dictionary is associated to each node,
    which is where we store the actual :class:`Operator`. This dictionary can
    be accessed by:

    .. code-block:: python

        pipeline.nodes[operator_idx]

    The :class:`Operator` is stored under the key 'operator':

    .. code-block:: python

        pipeline.nodes[operator_idx]['operator']

    We also store inputs to the operator under the key 'inputs':

    .. code-block:: python

        pipeline.nodes[operator_idx]['inputs']

    This key will only be generated when we start running the pipeline. Its
    value will be a list of length equal to the number of expected inputs to
    the :class:`Operator`, and will initially be filled with Nones. The outputs
    of earlier nodes will be placed in that list as they are generated, to
    store them until it is time to run that :class:`Operator`.

    Nodes can be added to the graph using the :meth:`networkx.DiGraph.add_node`
    method. The contents of the dictionary can also be appended as keyword
    arguments:

    .. code-block:: python

        op = Operator(...)
        pipeline.add_node(operator_idx, operator=op)

    Edges are stored as pairs, where the first entry in the pair is the node it
    emerges from and the second entry is the node it enters into. Again, in our
    case, those are the indices, not the :class:`Operator` s. The edges also
    have dictionaries associated to them, which can be accessed by:

    .. code-block:: python

        pipeline.edges[from_idx, to_idx]

    Since the inputs to an :class:`Operator` must be ordered, we keep track of
    which input a particular output is going to in that dictionary:

    .. code-block:: python

        pipeline.edges[from_idx, to_idx]['in_slot']

    Edges can be added similarly to nodes:

    .. code-block:: python

        pipeline.add_edge(from_idx, to_idx, in_slot=in_slot)
    '''
    # TODO: Add more error checking. A lot more.

    def __call__(self, *images) -> Union[List[np.ndarray], np.ndarray]:
        # Put images into buffers
        for loader, _images in zip(self.get_loaders(), images):
            loader.buff = [_images]

        # Run pipeline
        with _read_write_from_to_buffer():
            self.run()

        # Retrieve output
        out = [writer.buff.pop() for writer in self.get_writers()]

        return out[0] if (len(out) == 1) else out

    def add_operator(self, operator: Operator, idx: Optional[int] = None) \
            -> int:
        '''
        This adds an :class:`Operator` to the pipeline, and returns the
        operator's index.
        '''
        if idx is None:
            idx = max(self.nodes, default=-1) + 1
        if idx in self.nodes and 'operator' in self.nodes[idx]:
            raise ValueError(f'Duplicate indices: {idx}')
        self.add_node(idx, operator=operator)
        return idx

    @classmethod
    def chain(cls, *operators):
        '''
        This is a convenience function for building a :class:`Pipeline` where
        you have a list of :class:`Operator` s to be applied in sequence,
        without other changes.
        '''
        # First, instantiate the graph with a list of edges.
        edges = [(i, i + 1, {'in_slot': 0}) for i in range(len(operators) - 1)]
        pipe = cls(edges)
        # And add the operators.
        for node_idx, operator in enumerate(operators):
            pipe.nodes[node_idx]['operator'] = operator
        return pipe

    def get_loaders(self) -> List[Loader]:
        '''
        Returns a list of the :class:`Loader` s in the :class:`Pipeline`,
        ordered by index.
        '''
        idxs = [i for i in self.nodes
                if isinstance(self.nodes[i]['operator'], Loader)]
        return [self.nodes[i]['operator'] for i in sorted(idxs)]

    def get_writers(self) -> List[Writer]:
        '''
        Returns a list of the :class:`Writer` s in the :class:`Pipeline`,
        ordered by index.
        '''
        idxs = [i for i in self.nodes
                if isinstance(self.nodes[i]['operator'], Writer)]
        return [self.nodes[i]['operator'] for i in sorted(idxs)]

    @classmethod
    def load(cls, path: str):
        '''
        Loads a serialized :class:`Pipeline` from disk.

        Args:
            path (str): Path to serialized :class:`Pipeline`.
        '''
        with open(path, 'r') as in_file:
            graph = json.load(in_file)

        pipe = cls()

        # TODO: Is there a more efficient way to do this?

        for node in graph:
            op_cls = OPERATOR_REGISTRY.get(node['operator'].pop('class'))
            op = op_cls(**node['operator'])
            pipe.add_operator(op, idx=node['index'])
            for to_idx, details in node['edges']:
                pipe.add_edge(node['index'], to_idx, **details)

        return pipe

    def release_writers(self):
        '''
        Releases the writers. This should be called when writing video to disk,
        after you have written all the frames you need to write. Without this
        call, the video will not save properly. This method is called by the
        :meth:`run` method automatically when it's finished.
        '''
        for writer in self.get_writers():
            writer.release()

    def run(self) -> Dict:
        '''
        This runs the :class:`Pipeline` until it runs out of input images. It
        returns a dictionary of runtimes.
        '''
        start_time = time.perf_counter()
        op_times = Counter()

        # Add empty input slots
        for node in self.nodes.values():
            node['inputs'] = [None] * node['operator'].num_inputs

        # The topological sort returns an iterator over the node
        # indices ordered so that we will not reach any node until we
        # have already visited the nodes generating its inputs.
        op_idxs = list(nx.topological_sort(self))

        # TODO: Add check on matching lengths of Loaders

        # The Loaders will generate an OutOfInputs exception when they run out
        # of inputs. This gives us a convenient way to break out of the loops.
        with suppress(OutOfInputs):
            while True:
                # Go through the operators.
                for op_idx in op_idxs:
                    # Retrieve the dictionary corresponding to the operator
                    # index.
                    op: Dict = self.nodes[op_idx]
                    # Check the inputs have been filled in.
                    if any(x is None for x in op['inputs']):
                        raise RuntimeError(
                            f'Input missing from operator {op_idx}: {repr(op)}'
                            f'; pipeline graph must be incomplete.'
                        )
                    op_name = op['operator'].__class__.__name__
                    op_start_time = time.perf_counter()
                    # Run the operator on the inputs.
                    out = op['operator'](*op.pop('inputs'))
                    op_times[op_name] += (time.perf_counter() - op_start_time)

                    # Place the output into the corresponding input slots of
                    # operators downstream of this operator.
                    for _, down_op_idx in self.out_edges(op_idx):
                        in_slot = self.edges[op_idx, down_op_idx]['in_slot']
                        self.nodes[down_op_idx]['inputs'][in_slot] = out
                    # Reset the inputs to this operator.
                    op['inputs'] = [None] * op['operator'].num_inputs

        self.release_writers()

        total_runtime = time.perf_counter() - start_time
        op_times['pipeline'] = total_runtime - sum(op_times.values())

        return op_times

    def save(self, path: str):
        '''
        Serializes a :class:`Pipeline` to disk as a JSON file.

        Args:
            path (str): Path to serialize the :class:`Pipeline` to.
        '''
        out = []
        for idx in self.nodes:
            out.append({
                'index': idx,
                'operator': self.nodes[idx]['operator']._to_json(),
                'edges': list(self.succ[idx].items())
            })

        with open(path, 'w') as out_file:
            json.dump(out, out_file, indent=4)

    def set_all_paths(self, *paths):
        '''
        Sets all I/O paths.

        Args:
            paths: Each path should be a path to read from or write to, with
            the input paths preceding the output paths.
        '''
        loaders, writers = self.get_loaders(), self.get_writers()
        if len(paths) != len(loaders) + len(writers):
            raise ValueError(
                f'Incorrect number of paths: is {len(paths)}, should be '
                f'{len(loaders) + len(writers)}'
            )

        n_loaders = len(loaders)

        self.set_loader_paths(*paths[:n_loaders])
        self.set_writer_paths(*paths[n_loaders:])

    def set_batch_size(self, batch_size: int):
        '''
        Sets the batch size of the :class:`Loader` s.

        Args:
            batch_size (int): Number of frames to load per iteration.
        '''
        for loader in self.get_loaders():
            loader.batch_size = batch_size

    def set_loader_paths(self, *paths):
        '''
        Sets the paths for the loaders.
        '''
        loaders = self.get_loaders()
        if len(paths) != len(loaders):
            raise ValueError(
                f'Incorrect number of paths: is {len(paths)}, should be '
                f'{len(loaders)}'
            )

        for path, loader in zip(paths, loaders):
            loader.set_path(path)

    def set_writer_paths(self, *paths):
        '''
        Sets the paths for the writers.
        '''
        writers = self.get_writers()
        if len(paths) != len(writers):
            raise ValueError(
                f'Incorrect number of paths: is {len(paths)}, should be '
                f'{len(writers)}'
            )

        for path, writer in zip(paths, writers):
            writer.set_path(path)


load_pipeline = Pipeline.load



import os
from setuptools import setup
import shutil


if __name__ == '__main__':
    # Hypothetically, somebody might run this from somewhere other than the
    # root directory.
    root_path = os.path.dirname(os.path.abspath(__file__))

    # Read in requirements.txt
    req_path = os.path.join(root_path, 'requirements.txt')
    with open(req_path, 'r') as reqs_file:
        requirements = list(reqs_file.readlines())

    setup(
        name='video2vision',
        version='0.1',
        author='Hanley Color Lab',
        author_email='dhanley@gmu.edu',
        url='https://github.com/HanleyColorLab/video2vision',
        packages=['video2vision'],
        setup_requires=requirements,
        install_requires=requirements,
    )



'''
This file contains various operators that change themselves when they are used.
This includes, for example, :class:`AutoAlign`, which uses the first batch of
images it recieves to calculate an alignment, then reuses that alignment for
all of the remaining batches.
'''
from copy import copy
from typing import Dict, List, Optional, Tuple

import cv2
import numpy as np

from .elementwise import build_linearizer
from .operators import (
    ConcatenateOnBands,
    Operator,
    OPERATOR_REGISTRY,
    Warp
)
from .utils import (
    _coerce_to_4dim,
    _evaluate_ecc_for_warp,
    extract_samples,
    locate_aruco_markers
)

__all__ = ['AutoAlign', 'AutoLinearize', 'AutoOperator', 'AutoTemporalAlign']


class AutoOperator(Operator):
    '''
    An :class:`AutoOperator` is an :class:`Operator` that uses the first batch
    of data it receives to calculate some of its parameters.
    '''


@OPERATOR_REGISTRY.register
class AutoAlign(Warp, AutoOperator):
    '''
    The :class:`AutoAlign` takes in two images and attempts to find a
    homography operation that aligns them, then merges them on the bands. This
    requires a good coarse alignment to work. This supports two methods:
    finding tiepoints using ARUCO markers, and finding alignments using the ECC
    method. The ARUCO method is very fast, but requires that the ARUCO markers
    be present in the scene. The ECC method is very slow, but does not. For
    details on the ECC method, see:

    https://hal.inria.fr/file/index/docid/864385/filename/PAMI_2008_preprint.pdf

    Note that, while the ARUCO method is more robust to the initial alignment
    then the ECC method, it is not able to detect markers that have been
    flipped, and will fail if one image has been flipped compared to the other.
    '''
    num_inputs = 2

    def __init__(self, max_iterations: int = 5000, eps: float = 1e-4,
                 sampling_mode: int = cv2.INTER_LINEAR,
                 num_votes: int = 1, bands: Optional[List[List[int]]] = None,
                 mask: Optional[Tuple[int, int, int, int]] = None,
                 method: str = 'any', coe: Optional[np.ndarray] = None,
                 output_size: Optional[Tuple[int, int]] = None):
        '''
        Args:
            max_iterations (int): Maximum iterations to run the ECC algorithm.
            This is only used by the ECC method.

            eps (float): Tolerance for the change in ECC for terminating the
            algorithm. Be careful not to set this too low or it can keep
            running more or less forever and end up jumping into a bad local
            minimum. This is only used by the ECC method.

            sampling_mode (int): Type of interpolation to use.

            num_votes (int): Run the automatic alignment on this many frames
            and compare the results to pick the best alignment.
            bands (optional, list of list of int): Which bands to use. If not
            set, use all.

            mask (optional, 4-tuple of int): If provided, only use this region
            inside the source image to find the match. Should be provided as
            (left x, top y, right x, bottom y). This is only used by the ECC
            method.

            method (str): Method to use. Choices: 'any' (defaults to ARUCO
            markers if available), 'aruco', 'ecc'.

            coe (optional, :class:`numpy.ndarray`): The warp used in the
            alignment. This is ordinarily calculated from the first batch, but
            we allow it to be passed as an argument so that a fit
            :class:`video2vision.AutoAlign` can be saved to disk and reloaded.

            output_size (optional, pair of int): The size of the output image.
            This is ordinarily calculated from the first batch, but we allow it
            to be passed as an argument so that a fit
            :class:`video2vision.AutoAlign` can be saved to disk and reloaded.
        '''
        self.max_iterations = max_iterations
        self.eps = eps
        self.sampling_mode = sampling_mode
        self.num_votes = num_votes
        self.mask = mask
        if method not in {'any', 'aruco', 'ecc'}:
            raise ValueError(f'Did not recognize method {method}')
        self.method = method
        if coe is None:
            self.coe = self.output_size = None
        else:
            self.coe, self.output_size = np.array(coe), output_size
        self._concatenate = ConcatenateOnBands(bands)

    def apply(self, source: Dict, control: Dict) -> Dict:
        # coe will be None if we haven't found an alignment yet.
        if self.coe is not None:
            source = super().apply(source)
            return self._concatenate(source, control)
        else:
            source = self._find_alignment(source, control)
            return self._concatenate(source, control)

    def _find_alignment(self, source: Dict, control: Dict) -> Dict:
        '''
        This a) finds the alignment between the source and control and b) sets
        the internal properties to correspond to that alignment. It returns the
        warped images. We break this out as a separate method so it can be
        called by some subclasses.
        '''
        if self.method == 'any':
            try:
                return self._find_alignment_by_aruco(source, control)
            except RuntimeError:
                return self._find_alignment_by_ecc(source, control)
        elif self.method == 'aruco':
            return self._find_alignment_by_aruco(source, control)
        elif self.method == 'ecc':
            return self._find_alignment_by_ecc(source, control)
        else:
            raise ValueError(f'Did not recognize method {self.method}')

    def _find_alignment_by_aruco(self, source: Dict, control: Dict) -> Dict:
        '''
        Uses ARUCO markers to find the alignment.
        '''
        if self.output_size is None:
            self.output_size = control['image'].shape[:2][::-1]

        # These will raise RuntimeErrors if no markers are found.
        source_ts, source_pts, source_ids = locate_aruco_markers(source)
        control_ts, control_pts, control_ids = locate_aruco_markers(control)

        # Screen for frames with markers present in both
        joint_ts = sorted(set(source_ts) & set(control_ts))
        if not joint_ts:
            raise RuntimeError('No markers found in common frames')

        size = control['image'].shape[:2][::-1]
        # n tracks how many examples we've compared, so we don't exceed the
        # number of votes
        best_ecc, n = -float('inf'), 0
        # These need to be lists so we can use the index method
        source_ts, control_ts = list(source_ts), list(control_ts)

        for t in joint_ts:
            s_idx, c_idx = source_ts.index(t), control_ts.index(t)
            s_ids, c_ids = list(source_ids[s_idx]), list(control_ids[c_idx])
            if set(s_ids) != set(c_ids):
                continue

            # Find tie points
            src_pts = [source_pts[s_idx][s_ids.index(i)] for i in c_ids]
            src_pts = np.array(src_pts).reshape(-1, 2)
            con_pts = control_pts[c_idx].reshape(-1, 2)

            self.coe = Warp.build_from_tiepoints(src_pts, con_pts, size).coe
            warped_image = super().apply(source)

            # Test the alignment by calculating ECC on the entire image stack.
            # Skip the test if we're only doing this once.
            if self.num_votes > 1:
                ecc = _evaluate_ecc_for_warp(control, warped_image)
            else:
                ecc = 0

            if ecc > best_ecc:
                best_ecc, best_coe, best_warped = ecc, self.coe, warped_image

            n += 1
            if n >= self.num_votes:
                break

        if best_ecc == -float('inf'):
            raise RuntimeError('AutoAlign failed to find viable alignment')

        self.coe = best_coe

        return best_warped

    def _find_alignment_by_ecc(self, source: Dict, control: Dict) -> Dict:
        '''
        Uses the ECC method to find the alignment.
        '''
        if self.output_size is None:
            self.output_size = control['image'].shape[:2][::-1]

        # We could save some compute by pre-converting everything to grayscale.
        # But that would take up a lot of memory, recomputing the means is
        # cheap, and we're tighter on memory then compute.
        source_image, control_image = source['image'], control['image']
        if 'mask' in control and not control['mask'].all():
            raise NotImplementedError(
                'We do not yet support masks in the control image of an '
                'AutoAlign'
            )

        # Select the frames we'll use for the alignment. We try to keep them
        # roughly evenly spaced to avoid using frames that are too close
        # together in time and therefore too similar to generate different
        # warps.
        if source_image.ndim == 4:
            num_frames = control_image.shape[2]
            num_votes = min(self.num_votes, num_frames)
            ts = np.linspace(0, num_frames - 1, num_votes, dtype=np.int64)
            source_frames = [source_image[:, :, t, :] for t in ts]
            control_frames = [control_image[:, :, t, :] for t in ts]
        else:
            num_frames = num_votes = 1
            source_frames = [source_image]
            control_frames = [control_image]

        warp_matrix = np.eye(3, 3, dtype=np.float32)
        criteria = cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT
        self.output_size = control_frames[0].shape[:2][::-1]

        # Generate mask for ECC calculations
        if self.mask is not None:
            mask = np.zeros(source_frames[0].shape[:2], dtype=np.uint8)
            mask[self.mask[1]:self.mask[3], self.mask[0]:self.mask[2]] = 1
            if source.get('mask') is not None:
                mask &= source['mask']
        else:
            mask = source.get('mask')

        best_ecc = -float('inf')
        # Iterate through contrl frame, source frame pairs, calculating an
        # alignment for each one.
        for control_frame, source_frame in zip(control_frames, source_frames):
            # This can raise a cv2.error if it fails to converge. In that case,
            # we catch the error and try again with the next pair of frames.
            try:
                _, self.coe = cv2.findTransformECC(
                    # Need to specify the dtype because numpy defaults to
                    # float64 and cv2 needs float32 or uint8.
                    source_frame.mean(2, dtype=np.float32),
                    control_frame.mean(2, dtype=np.float32),
                    # cv2.findTransformECC modifies the warp_matrix in-place,
                    # so we need to clone it to avoid over-writing previous
                    # values.
                    warp_matrix.copy(),
                    cv2.MOTION_HOMOGRAPHY,
                    (criteria, self.max_iterations, self.eps),
                    mask,
                    # By default, the findTransformECC applies a Gaussian blur
                    # to the image prior to running. We can't turn it off, but
                    # we can set the kernel size to 1.
                    1,
                )

            except cv2.error:
                continue

            warped_image = super().apply(source)

            # Test the alignment by calculating ECC on the entire image stack.
            # Skip the test if we're only doing this once.
            if self.num_votes > 1:
                ecc = _evaluate_ecc_for_warp(control, warped_image)
            else:
                ecc = 0

            if ecc > best_ecc:
                best_ecc, best_coe, best_warped = ecc, self.coe, warped_image

        if best_ecc == -float('inf'):
            raise RuntimeError('AutoAlign failed to find viable alignment')

        self.coe = best_coe

        return best_warped

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'coe': (None if self.coe is None else self.coe.tolist()),
            'output_size': self.output_size,
            'max_iterations': self.max_iterations,
            'eps': self.eps,
            'sampling_mode': self.sampling_mode,
            'num_votes': self.num_votes,
            'mask': self.mask,
            'bands': self._concatenate.bands,
            'method': self.method,
        }


@OPERATOR_REGISTRY.register
class AutoLinearize(AutoOperator):
    '''
    The :class:`AutoLinearize` uses ARUCO markers to localize samples of known
    reflectivity in an image, then calculates a linearization to correct post-
    processing by the camera.

    Note that the ARUCO marker detection procedure is robust to rotation, but
    *not* to flipping, and will fail if the image has been flipped compared to
    the expected orientation.
    '''
    def __init__(self, marker_ids: List[int], marker_points: np.ndarray,
                 sample_points: np.ndarray, expected_values: np.ndarray,
                 method: str = 'poly', order: int = 3, sample_width: int = 25,
                 linearization: Optional[str] = None):
        '''
        Args:
            marker_ids (list of int): IDs of the fiducial markers.

            marker_points (:class:`numpy.ndarray`): Locations of the marker
            corners in a reference image.

            sample_points (:class:`numpy.ndarray`): Locations of the centers of
            the samples in a reference image.

            expected_values (:class:`numpy.ndarray`): Expected pixel magnitudes
            of the samples in a reference image.

            method (str): Method to use to fit the linearization. Choices:
            'poly', 'power'.

            order (int): Order of the polynomial to use for the linearization,
            if using a polynomial.

            sample_width (int): Width of the samples to extract.

            linearization (optional, str): If provided, this is a JSON string
            serializing the linearization operator. This is ordinarily
            calculated from the first batch, but we allow it to be passed as an
            argument so that a fit :class:`video2vision.AutoLinearize` can be
            saved to disk and reloaded.
        '''
        self.marker_ids = np.array(marker_ids)
        self.marker_points = np.array(marker_points)
        self.sample_points = np.array(sample_points)
        self.expected_values = np.array(expected_values)
        self.order = order
        self.sample_width = sample_width
        self.method = method
        if linearization is None:
            self.op = None
        else:
            op_cls = OPERATOR_REGISTRY.get(linearization.pop('class'))
            self.op = op_cls(**linearization)

    def apply(self, x: Dict) -> Dict:
        if self.op is not None:
            return self.op(x)

        with _coerce_to_4dim(x):
            sample_points, t = self._locate_samples(x)
            samples = extract_samples(
                x['image'][:, :, t, :], sample_points, self.sample_width
            )

        self.op = build_linearizer(
            samples,
            self.expected_values,
            method=self.method,
            order=self.order,
        )

        return self.op(x)

    def _locate_samples(self, image: Dict):
        '''
        Locates simple points in an image by locating fiducial marks.

        Args:
            image (Dict): Image to be searched.
        '''
        ts, corners = locate_aruco_markers(image, self.marker_ids)
        t, corners = ts[0], corners[0]

        # Using the corners as tie points, build a warp from the original image
        # to the image we've just received. corners will then be a 3-
        # dimensional np.ndarray in arrangement (Marker ID, Corner, X/Y).
        warp = Warp.build_from_tiepoints(
            self.marker_points.reshape(-1, 2),
            corners.reshape(-1, 2),
            # output_size doesn't matter, so give it a dummy
            (0, 0)
        )

        # Apply that warp to the sample points in the original image to obtain
        # sample points in the new image.
        return warp.apply_points(self.sample_points), t

    def _to_json(self) -> Dict:
        return {
            'class': self.__class__.__name__,
            'marker_ids': self.marker_ids.tolist(),
            'marker_points': self.marker_points.tolist(),
            'sample_points': self.sample_points.tolist(),
            'expected_values': self.expected_values.tolist(),
            'method': self.method,
            'order': self.order,
            'sample_width': self.sample_width,
            'linearization': self.op if self.op is None else self.op._to_json()
        }


@OPERATOR_REGISTRY.register
class AutoTemporalAlign(AutoAlign, AutoOperator):
    '''
    The :class:`AutoTemporalAlign` takes in two videos and uses the ECC method
    to find both a temporal shift and a homography operation that aligns them.
    It then replaces itself with a :class:`TemporalShift` and :class:`Warp`
    operator that implements that alignment. It expects its 0th input to
    provide the video to be aligned - the source video - and the 1st input to
    provide the video to align to - the control video.

    This requires a good coarse alignment to work.
    '''
    def __init__(self, time_shift_range: Tuple[int, int],
                 max_iterations: int = 5000, eps: float = 1e-4,
                 sampling_mode: int = cv2.INTER_LINEAR,
                 num_votes: int = 1, bands: Optional[List[List[int]]] = None,
                 mask: Optional[Tuple[int, int, int, int]] = None,
                 method: str = 'any', coe: Optional[np.ndarray] = None,
                 output_size: Optional[Tuple[int, int]] = None,
                 time_shift: Optional[int] = None):
        '''
        Args:
            time_shift_range (tuple of int): Range of temporal shifts to
            evaluate.

            max_iterations (int): Maximum iterations to run the ECC algorithm.
            This is only used by the ECC method.

            eps (float): Tolerance for the change in ECC for terminating the
            algorithm. Be careful not to set this too low or it can keep
            running more or less forever and end up jumping into a bad local
            minimum. This is only used by the ECC method.

            sampling_mode (int): Type of interpolation to use.

            num_votes (int): Run the automatic alignment on this many frames
            and compare the results to pick the best alignment.
            bands (optional, list of list of int): Which bands to use. If not
            set, use all.

            mask (optional, 4-tuple of int): If provided, only use this region
            inside the source image to find the match. Should be provided as
            (left x, top y, right x, bottom y). This is only used by the ECC
            method.

            method (str): Method to use. Choices: 'any' (defaults to ARUCO
            markers if available), 'aruco', 'ecc'.

            coe (optional, :class:`numpy.ndarray`): The warp used in the
            alignment. This is ordinarily calculated from the first batch, but
            we allow it to be passed as an argument so that a fit
            :class:`video2vision.AutoTemporalAlign` can be saved to disk and
            reloaded.

            output_size (optional, pair of int): The size of the output image.
            This is ordinarily calculated from the first batch, but we allow it
            to be passed as an argument so that a fit
            :class:`video2vision.AutoTemporalAlign` can be saved to disk and
            reloaded.

            time_shift (optional, int): The amount of the time shift. This is
            ordinarily calculated from the first batch, but we allow it to be
            passed as an argument so that a fit
            :class:`video2vision.AutoTemporalAlign` can be saved to disk and
            reloaded.
        '''
        super().__init__(
            max_iterations=max_iterations, eps=eps,
            sampling_mode=sampling_mode, mask=mask, num_votes=num_votes,
            bands=bands, method=method, coe=coe, output_size=output_size,
        )
        self.time_shift_range = time_shift_range
        self.time_shift = time_shift
        self.buff = None
        self.buff_names = None

    def apply(self, source: Dict, control: Dict) -> Dict:
        if self.coe is not None:
            source, control = self._shift(source, control, self.time_shift)

            with _coerce_to_4dim(source), _coerce_to_4dim(control):
                n_f = min(source['image'].shape[2], control['image'].shape[2])
                source['image'] = source['image'][:, :, :n_f, :]
                control['image'] = control['image'][:, :, :n_f, :]

            return super().apply(source, control)

        best_ecc = -float('inf')

        # Get the number of frames available.
        with _coerce_to_4dim(source), _coerce_to_4dim(control):
            t = min(source['image'].shape[2], control['image'].shape[2])

        # Iterate through possible time shifts
        for time_shift in range(*self.time_shift_range):
            # If the time shift exceeds the number of frames available, then
            # skip.
            if abs(time_shift) >= t:
                continue

            shifted_source, shifted_control = self._shift(
                source, control, time_shift, no_buffer=True,
            )

            # This can raise a RuntimeError if we fail to find an alignment.
            try:
                warped_source = self._find_alignment(
                    shifted_source, shifted_control
                )
            except RuntimeError:
                continue

            ecc = _evaluate_ecc_for_warp(shifted_control, warped_source)
            if ecc > best_ecc:
                best_ecc, best_shift, best_coe = ecc, time_shift, self.coe

        if best_ecc == -float('inf'):
            raise RuntimeError(
                'AutoTemporalAlign failed to find viable alignment'
            )

        self.coe, self.time_shift = best_coe, best_shift

        return self.apply(source, control)

    def _shift(self, source: Dict, control: Dict, shift: int = 0,
               no_buffer: bool = False):
        '''
        Shifts the source and control. If positive, shifts source forward. If
        negative, shifts control backward. Unlike
        :class:`video2vision.TemporalShift`, this trims the video instead of
        padding it.

        Args:
            source (dict): Source image.

            control (dict): Control image.

            shift (int): Number of frames to shift.

            no_buffer (bool): If set, do not use the buffer to store the
            trimmed frames.
        '''
        source, control = copy(source), copy(control)

        if shift == 0:
            return source, control

        shifted, static = (source, control) if shift > 0 else (control, source)
        shift = abs(shift)

        with _coerce_to_4dim(shifted), _coerce_to_4dim(static):
            image, buff = np.split(shifted['image'], [-shift], axis=2)

            if no_buffer or self.buff is None:
                _, static['image'] = np.split(static['image'], [shift], axis=2)

            if not no_buffer:
                if self.buff is not None:
                    image = np.concatenate((self.buff, image), axis=2)
                self.buff = buff

            shifted['image'] = image

        if 'names' in static and (no_buffer or self.buff_names is None):
            static['names'] = static['names'][shift:]

        if 'names' in shifted:
            buff = shifted['names'][-shift:]
            shifted['names'] = shifted['names'][:-shift]
            if not no_buffer:
                if self.buff_names is not None:
                    shifted['names'] = self.buff_names + shifted['names']
                self.buff_names = buff

        return source, control

    def _to_json(self) -> Dict:
        rtn = super()._to_json()
        rtn['time_shift_range'] = self.time_shift_range
        rtn['time_shift'] = self.time_shift
        return rtn
